package parser;

import java.io.InputStream;

import error.LispException;
import parser.MalformedSExpressionException.*;
import scanner.LispScanner;
import sexpression.*;
import token.Eof;
import token.Token;

/**
 * Converts a stream of bytes into internal representations of Lisp
 * S-expressions.
 */
public class LispParser {

    private LispScanner scanner;
    private Token nextToken;
    private boolean isNextTokenStored;

    public LispParser(InputStream inputStream, String fileName) {
        scanner = new LispScanner(inputStream, fileName);
        nextToken = null;
        isNextTokenStored = false;
    }

    public boolean isEof() {
        if (!isNextTokenStored)
            storeNextToken();

        return nextToken instanceof Eof;
    }

    private void storeNextToken() {
        try {
            nextToken = scanner.getNextToken();
            isNextTokenStored = true;
        } catch (LispException e) {
            // This method should give the illusion of not actually reading
            // a token, so we ignore any exception. It will be thrown when we
            // return the next sExpression.
        }
    }

    public SExpression getNextSExpression() {
        if (!isNextTokenStored)
            nextToken = scanner.getNextToken();
        else
            isNextTokenStored = false;

        return sExpr();
    }

    // sExpr ::= NUMBER | IDENTIFIER | STRING | QUOTE_MARK sExpr |
    // LEFT_PAREN sExprTail
    private SExpression sExpr() {
        switch (nextToken.getType()) {
        case NUMBER:
            return new LispNumber(nextToken.getText());
        case IDENTIFIER:
            return new Symbol(nextToken.getText());
        case STRING:
            return new LispString(nextToken.getText());
        case QUOTE_MARK:
            nextToken = scanner.getNextToken();
            SExpression arg = sExpr();

            return new Cons(new Symbol("QUOTE"), new Cons(arg, Nil.getUniqueInstance()));
        case LEFT_PAREN:
            return sExprTail();
        case RIGHT_PAREN:
            throw new StartsWithRightParenthesisException(nextToken);
        case EOF:
            throw new EofEncounteredException(nextToken);
        default:
            throw new UnrecognizedTokenException(nextToken);
        }
    }

    // sExprTail ::= RIGHT_PAREN | sExpr sExprTail
    private SExpression sExprTail() {
        nextToken = scanner.getNextToken();

        switch (nextToken.getType()) {
        case RIGHT_PAREN:
            return Nil.getUniqueInstance();
        default:
            SExpression car = sExpr();
            SExpression cdr = sExprTail();

            return new Cons(car, cdr);
        }
    }

}
